﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.Serialization;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Configuration;
using System.ServiceModel.Description;
using System.ServiceModel.Dispatcher;
using System.ServiceModel.Web;
using System.Text;
using System.Xml;
using System.Xml.Linq;

namespace dk.nsi.seal
{
    public class NemidAssertionExchangeEndpointBehavior : IEndpointBehavior
    {
        ClientCredentials clientCredentials;

        public void AddBindingParameters(ServiceEndpoint endpoint, BindingParameterCollection bindingParameters)
        {
            clientCredentials = bindingParameters.OfType<ClientCredentials>().FirstOrDefault();
        }

        public void ApplyClientBehavior(ServiceEndpoint endpoint, ClientRuntime clientRuntime)
        {
            var collection = clientRuntime.MessageInspectors;
            collection.Add(new NemidMessageInspect { clientCredentials = clientCredentials });
        }

        public void ApplyDispatchBehavior(ServiceEndpoint endpoint, EndpointDispatcher endpointDispatcher)
        {
            throw new NotImplementedException();
        }

        public void Validate(ServiceEndpoint endpoint)
        {
        }
    }

    class NemidMessageInspect : IClientMessageInspector
    {
        public ClientCredentials clientCredentials;

        public void AfterReceiveReply(ref Message reply, object correlationState)
        {
            XmlDocument request = correlationState as XmlDocument;

            if (!reply.IsFault)
            {
                MessageBuffer msgbuf = reply.CreateBufferedCopy(int.MaxValue);
                var r = msgbuf.CreateMessage();

                var ms = new MemoryStream();
                using (var wr = XmlWriter.Create(ms)) r.WriteMessage(wr);
                ms.Position = 0;

                var signcheck = new SealSignedXml(ms);
                if (!signcheck.CheckEnvelopeSignature())
                {
                    throw new Exception("Response signature Error");
                }
                reply = msgbuf.CreateMessage();
            }
        }

        public object BeforeSendRequest(ref Message request, IClientChannel channel)
        {
            if (clientCredentials == null || clientCredentials.ClientCertificate.Certificate == null) throw new Exception("clientCredentials Certificate is missing");

            MessageBuffer msgbuf = request.CreateBufferedCopy(int.MaxValue);
            Message tmpMessage = msgbuf.CreateMessage();

            var xdoc = new XDocument();
            using (var wr = xdoc.CreateWriter())
            {
                tmpMessage.WriteMessage(wr);
            }

            //Hack 
            SealUtilities.CheckAndSetSamlDsPreFix(xdoc);

            //Fill header
            ns.SetMissingNamespaces(xdoc);
            var hd = xdoc.Root.Element(ns.xsoap + "Header");
            var ac = hd.Element(ns.xwsa2 + "Action") ?? hd.Element(ns.xwsa + "Action");
            hd.Add(new XElement(ns.xwsa + "Action", new XAttribute("mustUnderstand", "1"), new XAttribute(ns.xwsu + "Id", "action"), ac.Value),
                    new XElement(ns.xwsa + "MessageID", new XAttribute(ns.xwsu + "Id", "messageID"), "urn:uuid:" + Guid.NewGuid().ToString("D")),
                    new XElement(ns.xwsse + "Security", new XAttribute("mustUnderstand", "1"), new XAttribute(ns.xwsu + "Id", "security"),
                                    new XElement(ns.xwsu + "Timestamp", new XAttribute(ns.xwsu + "Id", "timestamp"),
                                        new XElement(ns.xwsu + "Created", DateTime.UtcNow.ToString("u").Replace(' ', 'T'))
                                        )
                                     )
                  );
            ac.Remove();
            xdoc.Root.Element(ns.xsoap + "Body").Add(new XAttribute(ns.xwsu + "Id", "body"));

            var signer = new SealSignedXml(xdoc);
            XmlDocument envelope = signer.Sign(clientCredentials.ClientCertificate.Certificate);

            //create new request Message from envelope
            using (var ms = new MemoryStream())
            {
                envelope.Save(ms);
                ms.Position = 0;
                using (var reader = XmlDictionaryReader.CreateTextReader(ms, XmlDictionaryReaderQuotas.Max))
                {
                    msgbuf = Message.CreateMessage(reader, int.MaxValue, request.Version).CreateBufferedCopy(int.MaxValue);
                    request = msgbuf.CreateMessage();
                }
            }
            return envelope;
        }
    }


    public class Seal2SamlEndpointBehavior : IEndpointBehavior
    {
        ClientCredentials clientCredentials;

        public void AddBindingParameters(ServiceEndpoint endpoint, BindingParameterCollection bindingParameters)
        {
            clientCredentials = bindingParameters.OfType<ClientCredentials>().FirstOrDefault() ??
            endpoint.Behaviors.OfType<ClientCredentials>().FirstOrDefault();
        }

        public void ApplyClientBehavior(ServiceEndpoint endpoint, ClientRuntime clientRuntime)
        {
            var collection = clientRuntime.MessageInspectors;
            collection.Add(new Seal2SamlInspector { clientCredentials = clientCredentials });
        }

        public void ApplyDispatchBehavior(ServiceEndpoint endpoint, EndpointDispatcher endpointDispatcher)
        {
            throw new NotImplementedException();
        }

        public void Validate(ServiceEndpoint endpoint)
        {
        }
    }

    public class Seal2SamlBehaviorExtentionElement : BehaviorExtensionElement
    {
        public override Type BehaviorType
        {
            get { return typeof(Seal2SamlEndpointBehavior); }
        }

        protected override object CreateBehavior()
        {
            return new Seal2SamlEndpointBehavior();
        }
    }

    class Seal2SamlInspector : IClientMessageInspector
    {
        public ClientCredentials clientCredentials;

        public void AfterReceiveReply(ref Message reply, object correlationState)
        {
            XmlDocument request = correlationState as XmlDocument;

            if (!reply.IsFault)
            {
                MessageBuffer msgbuf = reply.CreateBufferedCopy(int.MaxValue);
                var r = msgbuf.CreateMessage();

                var ms = new MemoryStream();
                using (var wr = XmlWriter.Create(ms)) r.WriteMessage(wr);
                ms.Position = 0;

                var signcheck = new SealSignedXml(ms);
                if (!signcheck.CheckEnvelopeSignature())
                {
                    throw new Exception("Response signature Error");
                }
                reply = msgbuf.CreateMessage();
            }
        }

        public object BeforeSendRequest(ref Message request, IClientChannel channel)
        {
            if (clientCredentials == null || clientCredentials.ClientCertificate.Certificate == null) throw new Exception("clientCredentials Certificate is missing");

            MessageBuffer msgbuf = request.CreateBufferedCopy(int.MaxValue);
            Message tmpMessage = msgbuf.CreateMessage();

            var xdoc = new XDocument();
            using (var wr = xdoc.CreateWriter())
            {
                tmpMessage.WriteMessage(wr);
            }

            //Hack 
            SealUtilities.CheckAndSetSamlDsPreFix(xdoc);

            //Fill header
            ns.SetMissingNamespaces(xdoc);
            var hd = xdoc.Root.Element(ns.xsoap + "Header");
            var ac = hd.Element(ns.xwsa2 + "Action") ?? hd.Element(ns.xwsa + "Action");
            hd.Add(new XElement(ns.xwsa + "Action", new XAttribute("mustUnderstand", "1"), new XAttribute(ns.xwsu + "Id", "action"), ac.Value),
                    new XElement(ns.xwsa + "MessageID", new XAttribute(ns.xwsu + "Id", "messageID"), "urn:uuid:" + Guid.NewGuid().ToString("D")),
                    new XElement(ns.xwsse + "Security", new XAttribute("mustUnderstand", "1"), new XAttribute(ns.xwsu + "Id", "security"),
                                    new XElement(ns.xwsu + "Timestamp", new XAttribute(ns.xwsu + "Id", "timestamp"),
                                        new XElement(ns.xwsu + "Created", DateTime.UtcNow.ToString("u").Replace(' ', 'T'))
                                        )
                                     )
                  );
            ac.Remove();
            xdoc.Root.Element(ns.xsoap + "Body").Add(new XAttribute(ns.xwsu + "Id", "body"));

            var signer = new SealSignedXml(xdoc);
            XmlDocument envelope = signer.Sign(clientCredentials.ClientCertificate.Certificate);

            //create new request Message from envelope
            using (var ms = new MemoryStream())
            {
                envelope.Save(ms);
                ms.Position = 0;
                using (var reader = XmlDictionaryReader.CreateTextReader(ms, XmlDictionaryReaderQuotas.Max))
                {
                    msgbuf = Message.CreateMessage(reader, int.MaxValue, request.Version).CreateBufferedCopy(int.MaxValue);
                    request = msgbuf.CreateMessage();
                }
            }
            return envelope;
        }
    }


    //SealEndpointBehavior 
    public class SealEndpointBehavior : IEndpointBehavior
    {
        public void AddBindingParameters(ServiceEndpoint endpoint, BindingParameterCollection bindingParameters)
        {
        }

        public void ApplyClientBehavior(ServiceEndpoint endpoint, ClientRuntime clientRuntime)
        {
            var collection = clientRuntime.MessageInspectors;
            collection.Add(new SealMessageInspect());
        }

        public void ApplyDispatchBehavior(ServiceEndpoint endpoint, EndpointDispatcher endpointDispatcher)
        {
            endpointDispatcher.DispatchRuntime.MessageInspectors.Add(new SealMessageInspect());
        }

        public void Validate(ServiceEndpoint endpoint)
        {
        }
    }

    public class SealBehaviorExtentionElement : BehaviorExtensionElement
    {
        public override Type BehaviorType
        {
            get { return typeof(SealEndpointBehavior); }
        }

        protected override object CreateBehavior()
        {
            return new SealEndpointBehavior();
        }
    }

    class SealMessageInspect : IDispatchMessageInspector, IClientMessageInspector
    {
        public object AfterReceiveRequest(ref Message request, IClientChannel channel, InstanceContext instanceContext)
        {
            var xdoc = new XDocument();
            using (var wr = xdoc.CreateWriter())
            {
                wr.WriteStartElement("DGWSInfo");
                request.Headers.WriteHeader(request.Headers.FindHeader("Security", ns.wsse), wr);
                request.Headers.WriteHeader(request.Headers.FindHeader("Header", ns.dgws), wr);
                wr.WriteEndElement();
            }

            var err = SealUtilities.ValidateSecurity(xdoc.Root.Descendants(ns.xwsse + "Security").FirstOrDefault());

            if (err != null)
            {
                xdoc.Root.Add(new XElement("Fault",
                                new XElement("reason", err.Item1),
                                new XElement("detail", err.Item2))
                        );

                throw new FaultException<string>(xdoc.ToString(SaveOptions.DisableFormatting), new FaultReason("requesterror"));
            }
            return xdoc;
        }

        public void BeforeSendReply(ref Message reply, object correlationState)
        {
            var reqxdoc = correlationState as XDocument;

            if (reply.IsFault) //replaces s:Server with Server in faultcode
            {
                MessageBuffer msgbuf = reply.CreateBufferedCopy(int.MaxValue);
                var r = msgbuf.CreateMessage();

                var xdoc = new XDocument();
                using (var wr = xdoc.CreateWriter())
                {
                    r.WriteMessage(wr);
                }

                if (xdoc.Descendants("faultstring").First().Value == "requesterror")
                {
                    var fc = xdoc.Descendants(ns.xdgws + "FaultCode").First();
                    reqxdoc = XDocument.Parse(fc.Value);
                    var xfault = reqxdoc.Root.Element("Fault");
                    xdoc.Descendants("faultstring").First().Value = xfault.Element("reason").Value;
                    fc.Value = xfault.Element("detail").Value;
                }

                xdoc.Descendants("faultcode").First().Value = "Server";
                using (var reader = xdoc.CreateReader())
                {
                    var rmsg = Message.CreateMessage(reader, int.MaxValue, reply.Version).CreateBufferedCopy(int.MaxValue);
                    reply = rmsg.CreateMessage();
                }
            }

            if (reqxdoc != null)
            {
                var msg = reqxdoc.Descendants(ns.xdgws + "MessageID").First();
                reqxdoc.Descendants(ns.xdgws + "Linking").First().Add(new XElement(ns.xdgws + "RequireNonRepudiationReceipt", msg.Value));
                msg.Value = Guid.NewGuid().ToString("D");

                reply.Headers.Add(new DgwsMessageHeader(new DgwsHeader(reqxdoc.Root.Element(ns.xdgws + "Header"))));
                reqxdoc.Descendants(ns.xdgws + "FlowStatus").First().Value = "flow_finalized_succesfully";
            }

        }

        public void AfterReceiveReply(ref Message reply, object correlationState)
        {
            if (reply.IsFault)
            {
                MessageBuffer msgbuf = reply.CreateBufferedCopy(int.MaxValue);
                Message tmpMessage = msgbuf.CreateMessage();

                var xdoc = new XDocument();
                using (var wr = xdoc.CreateWriter())
                {
                    tmpMessage.WriteBody(wr);
                }

                var q = from h in xdoc.Root.Elements(ns.xsoap + "Header")
                        from i in h.Elements(ns.xsosi + "implicitLoginHeader")
                        from r in i.Elements(ns.xsosi + "requestIdCardDigestForSigningResponse")
                        select new SosiGWLoginError()
                        {
                            DigestValue = r.Element(ns.xds + "DigestValue").Value,
                            BrowserUrl = r.Element(ns.xsosi + "BrowserUrl").Value
                        };

                var le = q.FirstOrDefault();
                var reason = xdoc.Root.Elements(ns.xsoap + "Body")
                                    .Elements(ns.xsoap + "Fault")
                                    .Elements("faultstring")
                                    .Select(v => v.Value).FirstOrDefault();
                if (le != null)
                    throw new FaultException<SosiGWLoginError>(le, new FaultReason(reason));

                var dgwsfc = xdoc.Root.Descendants(ns.xdgws + "FaultCode").FirstOrDefault();
                if (dgwsfc != null)
                {
                    var fs = xdoc.Root.Descendants("faultstring").First();
                    throw new FaultException<XElement>(dgwsfc, new FaultReason(fs.Value), new FaultCode("Server"));
                }

                reply = msgbuf.CreateMessage();
            }
        }

        public object BeforeSendRequest(ref Message request, IClientChannel channel)
        {
            var hdidx = request.Headers.FindHeader("Security", ns.wsse);
            if (hdidx == -1) throw new FaultException("Security header is missing");
            var hd = request.Headers[hdidx];

            if (!(hd is SealCardMessageHeader))
            {
                MessageBuffer msgbuf = request.CreateBufferedCopy(int.MaxValue);
                Message tmpMessage = msgbuf.CreateMessage();

                var xdoc = new XDocument();
                using (var wr = xdoc.CreateWriter())
                {
                    tmpMessage.WriteMessage(wr);
                }

                //Hack 
                SealUtilities.CheckAndSetSamlDsPreFix(xdoc);

                //create new request Message from envelope
                using (var ms = new MemoryStream())
                {
                    xdoc.Save(ms, SaveOptions.DisableFormatting);
                    ms.Position = 0;
                    using (var reader = XmlDictionaryReader.CreateTextReader(ms, XmlDictionaryReaderQuotas.Max))
                    {
                        msgbuf = Message.CreateMessage(reader, int.MaxValue, request.Version).CreateBufferedCopy(int.MaxValue);
                        request = msgbuf.CreateMessage();
                    }
                }
            }
            else
            {
                request.Headers.RemoveAt(hdidx);
                request.Headers.Insert(0, hd as SealCardMessageHeader);
            }
            return null;
        }
    }
}
